### Persefone.jl - a model of agricultural landscapes and ecosystems in Europe.
###
### These are the tests for the nature model (excluding individual species).
###

## Test species definition

# new species have to be defined within the Persefone module scope, otherwise things don't work
Persefone.eval(quote

@species Mermaid begin
    ageofmaturity = 2
    pesticidemortality = 1.0
end

@create Mermaid begin
    @debug "Created $(animalid(self))."
end

@phase Mermaid life begin
    @debug "$(animalid(self)) is swimming happily in its pond."
    @respond pesticide @kill(self.pesticidemortality, "poisoning")
    @respond harvesting @setphase(drought)
    if self.sex == female && length(@neighbours()) < 3 &&
        self.age >= self.ageofmaturity && @landcover() == water
        @reproduce()
    end
end

@phase Mermaid drought begin
    n = sum(1 for a in @neighbours())
    @debug "$(animalid(self)) is experiencing drought with $n neighbour(s)."
    @respond sowing @setphase(life)
end

@populate Mermaid begin
    birthphase = life
    initphase = life
    habitat = @habitat(@landcover() == water)
    pairs=true
end
               
end) # end eval

## Test sets

@testset "Habitat macros" begin
    # set up the testing landscape
    model = inittestmodel()
    model.landscape[6,6] = Pixel(Ps.agriculture, Ps.sand, 1, [], [], [])
    fp = Ps.FarmPlot(
        1, [(6,6)], 1, Ps.nosoildata,
        Ps.ALMaSS.CropState(
            croptype = model.crops["winter wheat"],
            phase = Ps.ALMaSS.janfirst,
        )
    )
    push!(model.farmplots, fp)
    push!(model.animals,
          Ps.Mermaid(1, Ps.male, (-1,-1), (3,3), Ps.life),
          Ps.Mermaid(2, Ps.female, (-1,-1), (4,4), Ps.life))
    # create a set of habitat descriptors
    h1 = @habitat(@landcover() == Ps.water)
    h2 = @habitat(@cropname() == "winter wheat" &&
                  @cropheight() < 2m) #XXX randomly chosen value
    h3 = @habitat(@distanceto(Ps.water) > 20m &&
                  @distancetoedge() <= 20m)
    #FIXME nested macros don't work properly, counting seems wrong
    h4 = @habitat(@countanimals(radius=10m) == 1) 
    # test the descriptors
    @test h1((6,4), model) == true
    @test h1((5,4), model) == false
    @test h2((6,6), model) == true
    @test h2((5,6), model) == false
    @test h3((3,3), model) == true
    @test h3((5,4), model) == false
    @test h3((6,1), model) == false
    @test_broken h4((2,2), model) == true
    @test_broken h4((3,3), model) == false
    @test_broken h4((1,1), model) == false
end

@testset "Species initialisation" begin
    model = inittestmodel()
    # initialisation parameters, set 1
    initparams1 = Ps.PopInitParams(birthphase = Ps.life, initphase = Ps.life, popsize = 10)
    @test_logs((:info, "Initialised 10 Mermaids."),
               Ps.initpopulation!(Ps.Mermaid, initparams1, model))
    @test all(a -> a.sex in (Ps.male, Ps.female), model.animals)
    Ps.killallanimals!(model)
    # initialisation parameters, set 2
    initparams2 = Ps.PopInitParams(birthphase = Ps.life, initphase = Ps.life,
                                   popsize = 8*8*3, asexual=true)
    @test_logs((:info, "Initialised 192 Mermaids."),
               Ps.initpopulation!(Ps.Mermaid, initparams2, model))
    @test Ps.countanimals((1,1), model) == Ps.countanimals((6,6), model) == 3
    @test all(a -> a.sex == Ps.hermaphrodite, model.animals)
    Ps.killallanimals!(model)
    # initialisation parameters, set 3
    initparams3 = Ps.PopInitParams(birthphase = Ps.life, initphase = Ps.life, pairs = true,
                                   habitat = @habitat(@landcover() == Ps.grass))
    @test_logs((:warn, "initpopulation!() called with popsize and indarea both <= 0"),
               (:info, "Initialised 36 Mermaids."),
               Ps.initpopulation!(Ps.Mermaid, initparams3, model))
    @test Ps.countanimals((2,2), model, radius=20m) == Ps.countanimals((5,3), model, radius=10m) == 0
    @test Ps.countanimals((5,5), model) == Ps.countanimals((6,6), model) == 2
    a1, a2 = Ps.nearby_animals((6,6), model, radius=0m)
    @test a1.sex != a2.sex
    Ps.killallanimals!(model)
    # initialisation parameters, set 4
    initparams4 = Ps.PopInitParams(birthphase = Ps.life, initphase = Ps.life, popsize = 10,
                                   habitat = @habitat(@landcover() == Ps.water &&
                                                      @countanimals() < 5))
    @test_logs((:warn, "There are not enough suitable locations for Mermaid in the landscape."),
               (:info, "Initialised 5 Mermaids."),
               Ps.initpopulation!(Ps.Mermaid, initparams4, model))
    @test Ps.countanimals((1,1), model, radius=40m) == 0
    @test Ps.countanimals((6,4), model) == 5
end

@testset "Species macros" begin
    model = inittestmodel()
    pond = (6,4)
    # test individual initialisation
    mermaid = Ps.Mermaid(1)
    @test typeof(mermaid) <: Animal
    @test typeof(Ps.life) <: Function
    @test typeof(mermaid.phase) <: Function
    @test_logs((:debug, "Created Mermaid 1."), min_level=Logging.Debug,
               Ps.create!(mermaid, Ps.withtestlogger(model)))
    # test population initialisation
    @test_logs((:warn, "initpopulation!() called with popsize and indarea both <= 0"),
               (:info, "Initialised 2 Mermaids."),
               Ps.initpopulation!("Mermaid", model))
    @test Ps.nagents(model) == 2
    @test Ps.countanimals((1,1), model, radius=40m) == 0
    @test Ps.countanimals(pond, model) == 2
    @test model.animals[1].age == 0
    # test event handling
    createevent!(model, [pond], Ps.harvesting)
    @test_logs((:debug, "Mermaid 1 is swimming happily in its pond."),
               (:debug, "Mermaid 2 is swimming happily in its pond."),
               min_level=Logging.Debug, match_mode=:any,
               stepsimulation!(Ps.withtestlogger(model)))
    @test model.animals[1].age == 1
    @test model.animals[2].phase == Ps.drought
    createevent!(model, [pond], Ps.sowing)
    @test_logs((:debug, "Mermaid 1 is experiencing drought with 1 neighbour(s)."),
               (:debug, "Mermaid 2 is experiencing drought with 1 neighbour(s)."),
               min_level=Logging.Debug, match_mode=:any,
               stepsimulation!(Ps.withtestlogger(model)))
    @test_logs((:debug, "Mermaid 1 is swimming happily in its pond."),
               (:debug, "Mermaid 2 is swimming happily in its pond."),
               (:debug, "Mermaid 2 has reproduced."),
               (:debug, "Created Mermaid 3."),
               min_level=Logging.Debug, match_mode=:any,
               stepsimulation!(Ps.withtestlogger(model)))
    @test Ps.countanimals(pond, model) == 3
    createevent!(model, [pond], Ps.pesticide)
    @test_logs((:debug, "Mermaid 1 is swimming happily in its pond."),
               (:debug, "Mermaid 1 has died from poisoning."),
               (:debug, "Mermaid 2 is swimming happily in its pond."),
               (:debug, "Mermaid 2 has died from poisoning."),
               (:debug, "Mermaid 3 is swimming happily in its pond."),
               (:debug, "Mermaid 3 has died from poisoning."),
               min_level=Logging.Debug, match_mode=:any,
               stepsimulation!(Ps.withtestlogger(model)))
    @test Ps.countanimals(pond, model) == 0
    # test @rand (this is done more easily outside of @species)
    @test typeof(@rand()) == Float64
    @test @rand([true, true])
    #TODO test movement macros
    #TODO test territory macros
end

#XXX I'm unsure whether to keep the insect submodel at all, so rather than
# fixing the tests I'll comment them out for now. (There were problems related to the units.)
# @testset "Insect submodel" begin
#     # create a set of pixels and dates for testing
#     model = inittestmodel()
#     date1 = Date("2023-05-08") # day 128 (season begin)
#     date2 = Date("2023-07-06") # day 187 (insect max)
#     date3 = Date("2023-09-27") # day 270 (season end)
#     p1 = Pixel(Ps.agriculture, 1)
#     p2 = Pixel(Ps.agriculture, 1, [Ps.pesticide], [], [])
#     p3 = Pixel(Ps.grass, 1)
#     p4 = Pixel(Ps.soil, 1, [Ps.fertiliser, Ps.pesticide], [], [])
#     p5 = Pixel(Ps.forest, 1)
#     p6 = Pixel(Ps.water, 1)
#     # check whether the model calculates the same numbers I did by hand
#     model.date = date1
#     #FIXME these tests are still broken because insectbiomass throws a unit error
#     println(Ps.insectbiomass(p1, model))
#     println(Ps.insectbiomass(p2, model))
#     println(Ps.insectbiomass(p3, model))
#     println(Ps.insectbiomass(p4, model))
#     println(Ps.insectbiomass(p5, model))
#     println(Ps.insectbiomass(p6, model))
#     @test Ps.insectbiomass(p1, model) ≈ 0.00411g/m² atol=0.0001g/m²
#     @test Ps.insectbiomass(p2, model) == 0.0g/m²
#     @test Ps.insectbiomass(p3, model) ≈ 0.15411g/m² atol=0.0001g/m²
#     @test Ps.insectbiomass(p4, model) ≈ 0.00411g/m² atol=0.0001g/m²
#     @test Ps.insectbiomass(p5, model) ≈ 0.30411g/m² atol=0.0001g/m²
#     @test Ps.insectbiomass(p6, model) == 0.0g/m²
#     model.date = date2
#     @test Ps.insectbiomass(p1, model) == 0.3g/m²
#     @test Ps.insectbiomass(p2, model) == 0.15g/m²
#     @test Ps.insectbiomass(p3, model) == 0.45g/m²
#     @test Ps.insectbiomass(p4, model) == 0.3g/m²
#     @test Ps.insectbiomass(p5, model) == 0.6g/m²
#     @test Ps.insectbiomass(p6, model) == 0.0g/m²
#     model.date = date3
#     @test Ps.insectbiomass(p1, model) == 0.0g/m²
#     @test Ps.insectbiomass(p2, model) == 0.0g/m²
#     @test Ps.insectbiomass(p3, model) == 0.0g/m²
#     @test Ps.insectbiomass(p4, model) == 0.0g/m²
#     @test Ps.insectbiomass(p5, model) ≈ 0.01443g/m² atol=0.0001g/m²
#     @test Ps.insectbiomass(p6, model) == 0.0g/m²
# end

#XXX I probably won't keep the energy submodel at all, so commenting out the tests for now
# @testset "Energy submodel" begin
#     # DEB data for the skylark (https://bio.vu.nl/thb/deb/deblab/add_my_pet/entries_web/Alauda_arvensis/Alauda_arvensis_res.html)
#     skylarkparams = Ps.DEBparameters(0.0, # F_m XXX is unknown
#                                      0.8, # y_EX Sibly et al. (2013)
#                                      0.8075, # y_VE
#                                      0.04761, # v
#                                      0.0, # J_ET XXX is unknown
#                                      4699,# J_EM XXX seems awfully high?
#                                      0.9886, # k
#                                      0.95,# k_R
#                                      0.0, # M_E0 XXX is unknown
#                                      3.4, # M_Hb
#                                      25   # M_Hp XXX this seems too low?
#                                      )
#     #TODO
                                  
# end

#FIXME need to be updated, or removed altogether (the skylark code is so integrated
# that it is quite hard to unit-test - POM-validation seems more appropriate here)
# @testset "Skylark submodel" begin
#     # set up a modified test landscape
#     model = inittestmodel()
#     #FIXME the tests here fail if @distancetoedge >= 50m in `skylarkhabitat`,
#     # as no individuals are initialised
#     for x in 1:6
#         for y in 5:6
#             model.landscape[x,y] = Pixel(Ps.agriculture)
#         end
#     end
#     # test migration
#     @test_logs((:info, "Initialised 2 Skylarks."),
#                (:debug, "Skylark 1 has migrated."),
#                (:debug, "Skylark 2 has migrated."),
#                min_level=Logging.Debug, match_mode=:any,
#                Ps.initpopulation!("Skylark", Ps.withtestlogger(model)))
#     @test length(model.animals) == 2
#     @test all(isnothing, model.animals)
#     @test length(model.migrants) == 2
#     @test model.migrants[1].first.sex != model.migrants[2].first.sex
#     for a in model.migrants
#         leave, arrive = a.first.migrationdates
#         @test leave[1] in (9, 10) || (leave[1] == 11 && leave[2] <= 15)
#         @test (arrive[1] == 2 && arrive[2] >= 15) || (arrive[1] == 3 && arrive[2] <= 15)
#     end
#     model.date = Date(year(model.date), 3, 17)
#     @test_logs((:debug, "Skylark 1 has returned."),
#                (:debug, "Skylark 2 has returned."),
#                min_level=Logging.Debug, match_mode=:any,
#                Ps.updatenature!(Ps.withtestlogger(model)))
#     #TODO
# end
